`timescale 1ns / 1ps

//////////////////////////////////////////////////////////////////////////////////
//
//	Code written by: Jeff Casarona
//  Date: 10/15/2014
//  Property of UMD ECE Tech Ops
//
//		This code will initialize the LCD Screen for the Nexys 2 FPGA
//		and then sit in a endless loop, where it will give control over
//		to the "write_driver.v" code. The "JA" inout port acts as a bi-
//		directional bus. This means that in any case where you are not
//		actively driving the JA pins with data, you should hold them to
//		high impedance. 
//
//////////////////////////////////////////////////////////////////////////////////

module LCD_DRIVER(input clk, input btn1, input btn2, input [7:0] switches, 
			input reset, output RS, output R_W, output e, inout [7:0] JA
    );

	parameter wait_20ms = 22'h3F4240;
	parameter wait_39us = 22'h79E;
	parameter wait_1_53ms = 22'h12AD4;

	// Parameters for the statemachine
	localparam START=3'b000, SET_FUNC=3'b001, 
					SET_DISP=3'b010, CLEAR_DISP=3'b011, MODE_SET=3'b111, SET_HOME=3'b110,
					END_INIT=3'b100;
	
	wire RS_int;
	wire R_W_int;
	wire e_out;
	
	//Write Driver: This enables the writing of address locations
	// and of character data to the LCD
	write_driver wr_drv (
    .clk(clk), 
    .switches(switches), 
    .data_flag(btn1), 
    .addr_flag(btn2), 
    .reset(reset), 
    .e(e_out), 
    .RS(RS_int), 
    .RW(R_W_int), 
    .data(JA)
    );

	reg [2:0] state, nxt_state;
	reg [21:0] count;
	reg clp_en_clk;
	
	//Standard FSM Setup Code
	always @ (posedge clk) begin
		if(reset)
			state<=START;
		else
			state<=nxt_state;
	end 
	
	//Do next state and output updates
	// these are asynchronous
	always @ (count, state, reset) begin
		if(reset) begin
			nxt_state = START;
		end
		else begin
		case(state)
			START:
				// Wait 80ms
				nxt_state = (count == wait_20ms) ? SET_FUNC : START;
			SET_FUNC:
				// Wait (some amount of)us
				nxt_state=(count == wait_39us) ? SET_DISP : SET_FUNC;
			SET_DISP: 
				// Wait (some amount of)us
				nxt_state=(count == wait_39us) ? CLEAR_DISP : SET_DISP;
			CLEAR_DISP:
				// Wait (some amount of)ms
				nxt_state=(count == wait_1_53ms) ? MODE_SET : CLEAR_DISP;
			MODE_SET:
				// Wait (some amount of)us
				nxt_state= (count == wait_39us) ? SET_HOME : MODE_SET;
			SET_HOME:
				// Wait (some amount of)ms
				nxt_state = (count == wait_1_53ms) ? END_INIT : SET_HOME;
			END_INIT:
				nxt_state = END_INIT;
			default: begin 
				nxt_state=START;
				end
		endcase
		end
	end
	

	
	
	// Do Counting and Enable line
	// Note: For enable line, wait 40 ns before setting high,
	//		and hold this value for 220 ns before setting low
	always @ (posedge clk) begin
	if (reset) begin
		clp_en_clk <= 0;
		count <= 0;
	end
	else begin
		case(state)
			START: begin
				count <= (count == wait_20ms) ? 22'h0 : count + 1; // 20ms
				clp_en_clk <= 0;
				end
			SET_FUNC: begin
				count <= (count == wait_39us) ? 22'h0 : count + 1; // 37us
				clp_en_clk <= (count > 22'h1E && count < 22'h3C ) ? 1 : 0;
				end
			SET_DISP: begin
				count <= (count == wait_39us) ? 22'h0 : count + 1; // 37us
				clp_en_clk <= (count > 21'h1E && count < 22'h3C ) ? 1 : 0;
				end
			CLEAR_DISP: begin
				count <= (count == wait_1_53ms) ? 22'h0 : count + 1; // 1.52ms
				clp_en_clk <= (count > 22'h1E && count < 22'h3C ) ? 1 : 0;
				end
			MODE_SET: begin
				count <= (count == wait_39us) ? 22'h0 : count + 1;
				clp_en_clk <= (count > 22'h1E && count < 22'h3C) ? 1 : 0;
				end
			SET_HOME: begin
				count <= (count == wait_1_53ms) ? 22'h0 : count + 1;
				clp_en_clk <= (count > 22'h1E && count < 22'h3C ) ? 1 : 0;
				end
			END_INIT: begin
				count <= count + 1;
			end
			default: begin
				count <= 22'h0;
				clp_en_clk <= 0;
				end
		endcase
	end
	end
	
	//Enable line needs to update with clp enable
	assign e = (state == END_INIT) ? e_out : clp_en_clk;
	//These commands are in the datasheet for their respective function
	// Then set the output of this to high impedance (Z)
	assign JA = (state == SET_FUNC) ? 8'h3C : (state == SET_DISP) ? 8'h0F :
			(state == CLEAR_DISP) ? 8'h01 : (state == MODE_SET) ? 8'h06 : 
			(state == SET_HOME) ? 8'h02 : 8'hZ;//8'h41;
	// Hold low until the end of the initialization, where we pass
	//  control over to the write_driver.v
	assign RS = (state == END_INIT) ? RS_int : 0;
	assign R_W = (state == END_INIT) ? R_W_int : 0;	
	//assign R_W = 0;
	

endmodule
