`timescale 1ns / 1ps

//////////////////////////////////////////////////////////////////////////////////
//
//	Code written by: Jeff Casarona
//  Date: 10/15/2014
//  Property of UMD ECE Tech Ops
//
// This code is the main address and data driver to write characters to
// the LCD screen. Commands are triggered by pressing either button 1 or
// button 2, which will either set the address to write a character, or 
// write data to the LCD screen. Operation of the board:
//
//  After startup, blinking cursor will show at Address 0x00 of the board.
//  From here, "btn0" will reset the board by clearing the display and
//  setting the cursor back to the "Home" position, which is 0x00.
//
//  Pressing "btn1" will write data to the LCD screen. A standard ASCII
//  character scheme is followed for this. By default, writing a character
//  will increment your cursor to the next location. This makes it easy to
//  write a long string of characters.
//  
//  Pressing "btn2" will write the address to DDRAM. This will move your 
//  cursor to the corresponding location on the LCD Screen. The LCD
//  screen has 16 characters on two rows. The visible address locations
//  range from 0x00 to 0x0F for the top row, and 0x40 to 0x4F for the 
//  bottom row.
//
//  All data is input from the switches. For example, after programming the
//  board, entering 0x41 on the switches and pressing "btn1" will print a 
//  single "A" character in the top left corner of the display, and move
//  the cursor to the right by one. 
//
//////////////////////////////////////////////////////////////////////////////////

module write_driver(input clk, input [7:0] switches, input data_flag, input addr_flag, input reset, 
		output e, output RS, output RW, inout [7:0] data
    );
	 
	 localparam START=3'b000, BUSY_A=3'b001, WR_ADDR=3'b010, BUSY_D=3'b011, WR_DATA=3'b100;
	 parameter MAX_COUNT = 10'h3FF;
	 parameter PE_E = 10'h1E;
	 parameter NE_E = 10'h3C;
	 parameter ZERO_COUNT = 10'h00;
	 
	 reg [11:0] sh_data_btn;
	 reg [11:0] sh_addr_btn;
	 reg [2:0] state, nxt_state;
	 reg [9:0] count;
	 reg clp_en_clk;
	 
	 // Code used to generate a debounced button so only one character is
	 // printed at a time.
	 wire pe_data_btn, pe_addr_btn;
	 
	 
	 always @ (posedge clk) begin
		if (reset)
			state <= START;
		else
			state <= nxt_state;
	 end
	 
//////////////////////////////////////////////////////////////////////////////////
//  
//  This is the state machine to write the address or data to the LCD screen.
//  This process is a 3 step process
//
//  1. Wait for a button press (btn1 or btn2) to occur. Any enable signal should
// 		should be fine here
//  2. Check the busy flag from the LCD to make sure that the screen is not
//			doing anything currently. Do not leave until the LCD screen is !busy
//  3. Write the data or address from the switches to the LCD screen
//
//////////////////////////////////////////////////////////////////////////////////
	 always @ (state, pe_addr_btn, pe_data_btn, data[7], clp_en_clk, reset, count) begin
		nxt_state = START;
		if(reset) begin
			nxt_state = START;
		end
		else begin
			case(state)
				START: nxt_state = (pe_addr_btn) ? BUSY_A : (pe_data_btn) ? BUSY_D : nxt_state;
				BUSY_A: nxt_state = (clp_en_clk && data[7] == 1'b0) ? WR_ADDR : BUSY_A;
				// Delay the write states to slow down the board. This is not necessary
				WR_ADDR: nxt_state = (count == MAX_COUNT) ? START : WR_ADDR;
				BUSY_D: nxt_state = (clp_en_clk && data[7] == 1'b0) ? WR_DATA : BUSY_D;
				WR_DATA: nxt_state = (count == MAX_COUNT) ? START : WR_DATA;
				default: nxt_state = START;
			endcase
		end
	 end
	 
	 always @ (posedge clk) begin
		if(reset) begin
			sh_data_btn <= 0;
			sh_addr_btn <= 0;
			count <= ZERO_COUNT;
			clp_en_clk <= 0;
		end
		else begin
			// Debouncing code for the button presses Use a shift register
			sh_data_btn <= {sh_data_btn[10:0], data_flag};
			sh_addr_btn <= {sh_addr_btn[10:0], addr_flag};
			case(state)
				START: begin
					count <= ZERO_COUNT;
					clp_en_clk <= 0;
				end
				BUSY_A: begin
					count <= count + 1;
					clp_en_clk <= (count > PE_E && count < NE_E ) ? 1 : 0;
				end
				WR_ADDR: begin
					count <= count + 1;
					clp_en_clk <= (count > PE_E && count < NE_E ) ? 1 : 0;
				end
				BUSY_D: begin
					count <= count + 1;
					clp_en_clk <= (count > PE_E && count < NE_E ) ? 1 : 0;
				end
				WR_DATA: begin
					count <= count + 1;
					clp_en_clk <= (count > PE_E && count < NE_E ) ? 1 : 0;
				end
				default: begin
					count <= ZERO_COUNT;
					clp_en_clk <= 0;
				end
			endcase
		end
	 end
	 
	 
	 assign e = clp_en_clk;
	 assign RS = (state == WR_DATA) ? 1 : 0;
	 assign RW = ((state == BUSY_A) || (state == BUSY_D)) ? 1 : 0;
	 assign data = (state == WR_DATA) ? switches : (state == WR_ADDR) ? {1'b1,switches[6:0]} : 8'bZ;
	 
	 // Debounce logic: If 12'b011111111111 is shown, then that means we have a 
	 //  stable 0->1 transition, and you should move to the next state
	 // This debounce logic is not perfect. In order to use the buttons
	 //  and have perfect accuracy, debounce register needs to be larger
	 assign pe_data_btn = (sh_data_btn == 12'h7FF);
	 assign pe_addr_btn = (sh_addr_btn == 12'h7FF);

endmodule
